#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#if !defined(WIN32)
#include <unistd.h>
#include "serial.h"
#else
#include "win32/serial_win.h"
#endif

#include "target.h"

#include "log.h"

static int parse_initrd_start(char *buf, struct line_info *info);
static int parse_kernel_offset(char *buf, struct line_info *info);
static int parse_read(char *buf, struct line_info *info);
static int parse_write(char *buf, struct line_info *info);
static int parse_putc(char *buf, struct line_info *info);
static int parse_or(char *buf, struct line_info *info);
static int parse_and(char *buf, struct line_info *info);
static int parse_if(char *buf, struct line_info *info);
static int parse_endif(char *buf, struct line_info *info);
static int parse_print(char *buf, struct line_info *info);
static int parse_printbuf(char *buf, struct line_info *info);
static int parse_sleep(char *buf, struct line_info *info);
static int parse_serial_baud(char *buf, struct line_info *info);

const struct script_cmd_info sc_cmd[]={
  {sl_INITRD_START ,"INITRD_START",parse_initrd_start},
  {sl_KERNEL_OFFSET,"KERNEL_OFFSET",parse_kernel_offset},
  {sl_READB        ,"READB"       ,parse_read        },
  {sl_READW        ,"READW"       ,parse_read        },
  {sl_READ         ,"READ"        ,parse_read        },
  {sl_WRITEB       ,"WRITEB"      ,parse_write       },
  {sl_WRITEW       ,"WRITEW"      ,parse_write       },
  {sl_WRITE        ,"WRITE"       ,parse_write       },
  {sl_PUTC         ,"PUTC"        ,parse_putc        },
  {sl_OR           ,"OR"          ,parse_or          },
  {sl_AND          ,"AND"         ,parse_and         },
  {sl_IF           ,"IF"          ,parse_if          },
  {sl_ENDIF        ,"ENDIF"       ,parse_endif       },
  {sl_PRINT        ,"PRINT"       ,parse_print       },
  {sl_PRINTBUF     ,"PRINTBUF"    ,parse_printbuf    },
  {sl_SLEEP        ,"SLEEP"       ,parse_sleep       },
  {sl_SERIAL_BAUD  ,"SERIAL_BAUD" ,parse_serial_baud },
  {0               ,NULL          ,NULL              }
};

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
int get_line(FILE *fp, struct line_info *info){
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static char *get_item(char *buf, char *itembuf, int itembuf_size){
  char *cur;
  char *tmp;
  int len;
  if(buf == NULL) return NULL;
  cur = buf;

  cur += strspn(cur, " \t");
  switch(cur[0]){
  case 0:
  case '\r':
  case '\n':
  case '#':
    return NULL;
  default:
    tmp = strpbrk(cur, " \t\r\n");
    if(tmp == NULL){
      return NULL;
    }
    len = (int)((unsigned long)tmp - (unsigned long)cur);
    if(len > itembuf_size - 1){
      return NULL;
    }
    memset(itembuf, 0, itembuf_size);
    memcpy(itembuf, cur, len);
    cur = tmp;
  }
  return cur;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static char *get_message(char *buf, char *msgbuf, int msgbuf_size){
  char *cur;
  char *tmp;
  int len;

  if(buf == NULL) return NULL;
  cur = buf;

  cur += strspn(cur," \t");
  switch(cur[0]){
  case 0:
  case '\r':
  case '\n':
  case '#':
    return NULL;
  default:
    tmp = strpbrk(cur, "#\r\n");
    if(tmp == NULL){
      return NULL;
    }
    len = (int)((unsigned long)tmp - (unsigned long)cur);
    if(len > msgbuf_size -1){
      return NULL;
    }
    memset(msgbuf, 0, msgbuf_size);
    memcpy(msgbuf, cur, len);
  }
  return cur;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_initrd_start(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }

  arglen = strlen(arg);
  addr = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (addr == ULONG_MAX)
     ){
    return -1;
  }
  
  info->addr = addr;
  return 0;
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_kernel_offset(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }

  arglen = strlen(arg);
  addr = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (addr == ULONG_MAX)
     ){
    return -1;
  }
  
  info->addr = addr;
  return 0;
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_read(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;
  int i;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  //Internal Buffer detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  for(i=0; i<4; i++){
    char bufstr[16];
    sprintf(bufstr, "BUF%d", i+1);
    if(strcasecmp(bufstr, arg) == 0){
      info->internal_buf = i+1;
      break;
    }
  }
  if(info->internal_buf == 0) return -1;

  //Read Address detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  arglen = strlen(arg);
  addr = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (addr == ULONG_MAX)
     ){
    return -1;
  }
  info->addr = addr;
  return 0;
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_write(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;
  unsigned long value;
  int i;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  //Write Address detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  arglen = strlen(arg);
  addr = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (addr == ULONG_MAX)
     ){
    return -1;
  }
  info->addr = addr;

  //Value or Intenal Buffer detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  for(i=0; i<4; i++){
    char bufstr[16];
    sprintf(bufstr, "BUF%d", i+1);
    if(strcasecmp(bufstr, arg) == 0){
      info->internal_buf = i+1;
      return 0;
    }
  }
  arglen = strlen(arg);
  value = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (value == ULONG_MAX)
     ){
    return -1;
  }
  info->value = value;
  
  return 0;
}
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_putc(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long value;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }

  arglen = strlen(arg);
  value = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (value == ULONG_MAX)
     ){
    return -1;
  }
  
  info->value = value;
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_or(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;
  unsigned long value;
  int i;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  //Write Address or Internal Buffer detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  for(i=0; i<4; i++){
    char bufstr[16];
    sprintf(bufstr, "BUF%d", i+1);
    if(strcasecmp(bufstr, arg) == 0){
      info->internal_buf = i+1;
      break;
    }
  }
  if(info->internal_buf == 0){
    arglen = strlen(arg);
    addr = strtoul(arg, &endptr, 16);
    if(
       (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
       (addr == ULONG_MAX)
       ){
      return -1;
    }
    info->addr = addr;
  }

  //Value detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  arglen = strlen(arg);
  value = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (value == ULONG_MAX)
     ){
    return -1;
  }
  info->value = value;
  
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_and(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  unsigned long addr;
  unsigned long value;
  int i;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  //Write Address or Internal Buffer detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  for(i=0; i<4; i++){
    char bufstr[16];
    sprintf(bufstr, "BUF%d", i+1);
    if(strcasecmp(bufstr, arg) == 0){
      info->internal_buf = i+1;
      break;
    }
  }
  if(info->internal_buf == 0){
    arglen = strlen(arg);
    addr = strtoul(arg, &endptr, 16);
    if(
       (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
       (addr == ULONG_MAX)
       ){
      return -1;
    }
    info->addr = addr;
  }

  //Value detect
  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }
  arglen = strlen(arg);
  value = strtoul(arg, &endptr, 16);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (value == ULONG_MAX)
     ){
    return -1;
  }
  info->value = value;
  
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_if(char *buf, struct line_info *info){
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_endif(char *buf, struct line_info *info){
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_print(char *buf, struct line_info *info){
  char *cur;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_message(buf, info->buf, LINEINFO_BUFSIZE);
  if(cur == NULL){
    return -1;
  }
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_printbuf(char *buf, struct line_info *info){
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_sleep(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  int sleep;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }

  arglen = strlen(arg);
  sleep = strtoul(arg, &endptr, 10);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (sleep == ULONG_MAX)
     ){
    return -1;
  }
  
  info->value = sleep;
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_serial_baud(char *buf, struct line_info *info){
  char arg[256];
  char *cur;
  int value;

  char *endptr;
  int arglen;

  if(buf == NULL) return -1;
  cur = buf;

  cur = get_item(cur, arg, 256);
  if(cur == NULL){
    return -1;
  }

  arglen = strlen(arg);
  value = strtoul(arg, &endptr, 10);
  if(
     (endptr != (char *)((unsigned long)arg + (unsigned long)arglen)) ||
     (value == ULONG_MAX)
     ){
    return -1;
  }
  
  info->value = value;
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
static int parse_script_line(char *buf, struct line_info *info){
  char str[256];
  char label[64];
  char *cur;
  int ret;
  int i;

  if(buf == NULL) return -1;
  cur = buf;

  cur += strspn(cur, " \t");
  switch(cur[0]){
  case 0:
  case '\r':
  case '\n':
  case '#':
    info->label = 0;//comment
    return 0;
  default:
    sscanf(cur,"%s",label);
    cur = strpbrk(cur, " \t\r\n");
    if(cur == NULL){
      sprintf(str, "unknown line: %s\n",buf);
      message(str);
      return -1;
    }
  }
  
  for(i=0, info->label=0;; i++){
    if(sc_cmd[i].label == 0){
      sprintf(str, "unknown label: %s\n",label);
      message(str);
      return -1;
    }
    
    if(strcasecmp(label, sc_cmd[i].name) == 0){
      //label detect
      info->label = sc_cmd[i].label;
      break;
    }
  }
  
  if(sc_cmd[i].func == NULL){
    sprintf(str, "Label: [%s] function not implemented.\n",label);
    message(str);
    return -1;
  }
  ret = sc_cmd[i].func(cur, info);
  if(ret == -1){
    sprintf(str, "unknown format: %s",buf);
    message(str);
    return -1;
  }

  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
int execute_script(FILE *fp, int fake){
  char str[256];
  char buf[256];
  struct line_info info;
  int ret;
  static unsigned long internal_buf[4];

  memset(&internal_buf, 0, sizeof(unsigned long) * 4);

  while(feof(fp)==0){
    memset(&info, 0, sizeof(struct line_info));
    memset(buf, 0, 256);
    fgets(buf,256,fp);

    ret = parse_script_line(buf, &info);
    if(ret == -1) break;

    switch(info.label){
    case 0:
      continue;
    case sl_INITRD_START:
      break;
    case sl_READB:
	internal_buf[info.internal_buf] = target_read_byte(info.addr);
	break;
    case sl_READW:
    case sl_READ:
      //printf("read: BUF%d = 0x%08lx\n",info.internal_buf,info.addr);
      internal_buf[info.internal_buf] = target_read_word(info.addr);
      break;
    case sl_WRITEB:
      if(info.internal_buf != 0){
	//printf("write: 0x%08lx = BUF%d\n",info.addr,info.internal_buf);
	target_write_byte(info.addr, (unsigned char)internal_buf[info.internal_buf]);
      }else{
	//printf("write: 0x%08lx = 0x%08lx\n",info.addr,info.value);
	target_write_byte(info.addr, (unsigned char)info.value);
      }
      break;
    case sl_WRITEW:
    case sl_WRITE:
      if(info.internal_buf != 0){
	//printf("write: 0x%08lx = BUF%d\n",info.addr,info.internal_buf);
	target_write_word(info.addr, internal_buf[info.internal_buf]);
      }else{
	//printf("write: 0x%08lx = 0x%08lx\n",info.addr,info.value);
	target_write_word(info.addr, info.value);
      }
      break;
    case sl_PUTC:
      //printf("putc: %02x\n",(unsigned char)info.value);
      put_char((unsigned char)info.value);
      break;
    case sl_OR:
      if(info.internal_buf != 0){
	//printf("or: BUF%d |= 0x%08lx\n",info.internal_buf,info.value);
	internal_buf[info.internal_buf] |= info.value;
      }else{
	unsigned long tmp;
	tmp = target_read_word(info.addr);
	target_write_word(info.addr, tmp | info.value);
	//printf("or: 0x%08lx |= 0x%08lx\n",info.addr,info.value);	
      }
      break;
    case sl_AND:
      if(info.internal_buf != 0){
	//printf("or: BUF%d |= 0x%08lx\n",info.internal_buf,info.value);
	internal_buf[info.internal_buf] &= info.value;
      }else{
	unsigned long tmp;
	tmp = target_read_word(info.addr);
	target_write_word(info.addr, tmp & info.value);
	//printf("or: 0x%08lx |= 0x%08lx\n",info.addr,info.value);	
      }
      break;
    case sl_IF:
    case sl_ENDIF:
      break;
    case sl_PRINT:
      //printf("print message: %s\n",info.buf);
      sprintf(str, "%s\n",info.buf);
      message(str);
      break;
    case sl_PRINTBUF:
      break;
    case sl_SLEEP:
      if(fake == 0){
	int sec, usec;
	if(info.value){
	  sec = info.value / 1000;
	  usec = (info.value % 1000) * 1000;
	  if(sec) sleep(sec);
	  if(usec) usleep(usec);
	}
	//printf("sleep: %ld ms\n",info.value);	
      }
      break;
    case sl_SERIAL_BAUD:
      sleep(1);
      serial_baud(info.value);
      break;
    }
  }
  return 0;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
//+ 
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
unsigned long get_initrd_start(FILE *fp){
  char buf[256];
  struct line_info info;
  int ret;

  while(feof(fp)==0){
    memset(&info, 0, sizeof(struct line_info));
    memset(buf, 0, 256);
    fgets(buf,256,fp);

    ret = parse_script_line(buf, &info);
    if(ret == -1) break;

    switch(info.label){
    case sl_INITRD_START:
      return info.addr;
    default:
      continue;
    }
  }
  return -1;
}
unsigned long get_kernel_offset(FILE *fp){
  char buf[256];
  struct line_info info;
  int ret;

  while(feof(fp)==0){
    memset(&info, 0, sizeof(struct line_info));
    memset(buf, 0, 256);
    fgets(buf,256,fp);

    ret = parse_script_line(buf, &info);
    if(ret == -1) break;

    switch(info.label){
    case sl_KERNEL_OFFSET:
      return info.addr;
    default:
      continue;
    }
  }
  return -1;
}